(* Copyright (c) 2021 CIIRC (Czech Institute of Informatics, Robotics and Cybernetics) / CTU (Czech Technical University) *)

open Syntax
open Parser
open Sha256
open Hash

(*** The hash ids for the following will be added as we bootstrap, allowing us to make sense of more term syntax as we go. ***)
let fal : string ref = ref "";;
let fale : string ref = ref "";;
let eqPoly : string ref = ref "";;
let eqsymPoly : string ref = ref "";;
let eqsymPolyknown : bool ref = ref false;;
let eqSet : string ref = ref "";;
let conj : string ref = ref "";;
let disj : string ref = ref "";;
let expoly : string ref = ref "";;
let expolyI : string ref = ref "";;
let expolyIknown : bool ref = ref false;;
let setIn : string option ref = ref None;;
let setSubeq : string option ref = ref None;;
let setPow : string option ref = ref None;;
let setimplop : string option ref = ref None;;
let setlam : string option ref = ref None;;
let replop : string option ref = ref None;;
let sepop : string option ref = ref None;;
let replsepop : string option ref = ref None;;

let setenuml : string option list ref = ref [];;
let setenumadj : string option ref = ref None;;

let seqcons : string option ref = ref None;;

let rec set_setenuml_r i h sel =
  if i > 0 then
    begin
      match sel with
      | [] -> None::set_setenuml_r (i-1) h []
      | (z::ser) -> z::set_setenuml_r (i-1) h ser
    end
  else
    begin
      match sel with
      | [] -> [Some h]
      | (_::ser) -> Some h::ser
    end

let set_setenuml_n i h =
  setenuml := set_setenuml_r i h !setenuml

let nat0 : string option ref = ref None
let natS : string option ref = ref None

(*** Hash id for polymorphic if operator: ***)
let ifop : string option ref = ref None
(*** Hash id for monomorphic if operator: ***)
let ifopset : string option ref = ref None

let megacontra : string option ref = ref None;;

exception SearchLimit

let megasgdelta : (string,ptm) Hashtbl.t ref = ref (Hashtbl.create 10)

let rec local_hyp hyps g i =
  match hyps with
  | p::hypsr ->
     if conv p g !megasgdelta [] = None then
       local_hyp hypsr g (i+1)
     else
       i
  | [] -> raise Not_found

let basic_megaauto lm salt cx hyps g =
  if lm <= 0 then raise SearchLimit;
  try
    let i = local_hyp hyps g 0 in
    Hyp(i)
  with
  | Not_found ->
     let salt2 = hashtag salt 23l in
     let (s0,s1,s2,s3,s4,s5,s6,s7) = salt2 in
     ignore (s0,s1,s2,s3,s4,s5,s6,s7); (* real automation is unwritten, but when it is these salt int32s should guide it *)
     match g with
     | Imp(p,g) ->
        raise SearchLimit
     | _ ->
        raise SearchLimit

let rec basic_megaauto_search salt cx hyps g =
  let (s0,s1,s2,s3,s4,s5,s6,s7) = salt in
  try
    let d = basic_megaauto 1000 salt cx hyps g in
    let h = md256_hexstring salt in
    let n = ref 63 in
    while (!n >= 0 && h.[!n] = '0') do
      decr n
    done;
    Printf.printf "Line %d Char %d: Use ?%s?\n" !lineno !charno (String.sub h 0 (!n+1)); flush stdout;
    d
  with
  | SearchLimit ->
     if s0 = -1l then
       if s1 = -1l then
         if s2 = -1l then
           basic_megaauto_search (s0,s1,s2,Int32.add s3 1l,s4,s5,s6,s7) cx hyps g
         else
           basic_megaauto_search (s0,s1,Int32.add s2 1l,s3,s4,s5,s6,s7) cx hyps g
       else
         basic_megaauto_search (s0,Int32.add s1 1l,s2,s3,s4,s5,s6,s7) cx hyps g
     else
       basic_megaauto_search (Int32.add s0 1l,s1,s2,s3,s4,s5,s6,s7) cx hyps g
    
let megaauto pre ops cx hyps g =
  if pre = "" then
    if not (ops = "") && ops.[0] = '*' then (** search for an appropriate salt **)
      let n = String.length ops - 1 in
      let h =
        if n = 64 then
          String.sub ops 1 64
        else if n > 64 then
          raise (Failure (Printf.sprintf "salt for megaauto too long \"%s\"" ops))
        else
          let sbuf = Buffer.create 64 in
          for i = 1 to n do Buffer.add_char sbuf (ops.[i]) done;
          for i = n to 63 do Buffer.add_char sbuf '0' done;
          Buffer.contents sbuf
      in
      let salt = hexstring_md256 h in
      Printf.printf "Line %d Char %d: Trying Basic Megaauto Search (will likely hang until successful or killed)\n" !lineno !charno; flush stdout;
      basic_megaauto_search salt cx hyps g
    else
      let n = String.length ops in
      let h =
        if n = 64 then
          ops
        else if n > 64 then
          raise (Failure (Printf.sprintf "salt for megaauto too long \"%s\"" ops))
        else
          let sbuf = Buffer.create 64 in
          Buffer.add_string sbuf ops;
          for i = n to 63 do Buffer.add_char sbuf '0' done;
          Buffer.contents sbuf
      in
      let salt = hexstring_md256 h in
      basic_megaauto 1000 salt cx hyps g
  else
    raise (Failure (Printf.sprintf "unknown prefix \"%s\" for megaauto" pre))
